package ru.bitel.bgbilling.modules.inet.dyn.device.terminal;

import bitel.billing.server.util.mikrotik.MikrotikApiSession;
import bitel.billing.server.util.ssh.SSHSession;
import bitel.billing.server.util.telnet.TelnetSession;
import org.apache.log4j.Logger;
import ru.bitel.common.ParameterMap;
import ru.bitel.common.Utils;
import ru.bitel.common.io.IOUtils;
import ru.bitel.oss.systems.inventory.resource.common.bean.Device;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.URI;
import java.util.List;
import java.util.Scanner;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.regex.Pattern;

public abstract class TerminalSession
	implements AutoCloseable
{
	private static final Logger logger = Logger.getLogger( TerminalSession.class );

	protected String username;
	protected String password;

	/**
	 * Не ожидаем ответа
	 */
	public static int NOWAIT = 1;

	protected long timeout = 60000;
	protected long readWait = 50;

	protected String endSequence;
	protected String exitCommand = "exit";
	
	protected String sourceHost;
	protected int sourcePort;

	public String getUsername()
	{
		return username;
	}

	public void setUsername( String username )
	{
		this.username = username;
	}

	public String getPassword()
	{
		return password;
	}

	public void setPassword( String password )
	{
		this.password = password;
	}

	public void setTimeout( long timeout )
	{
		this.timeout = timeout;
	}

	public void setReadWait( long readWait )
	{
		this.readWait = readWait;
	}

	public void setEndSequence( String endSequence )
	{
		this.endSequence = endSequence;
	}

	public void setExitCommand( String exitCommand )
	{
		this.exitCommand = exitCommand;
	}

	public String getSourceHost()
	{
		return sourceHost;
	}

	public void setSourceHost( String sourceHost )
	{
		this.sourceHost = sourceHost;
	}

	public int getSourcePort()
	{
		return sourcePort;
	}

	public void setSourcePort( int sourcePort )
	{
		this.sourcePort = sourcePort;
	}

	public abstract void connect( String username, String password )
		throws Exception;

	public final void connect()
		throws Exception
	{
		connect( username, password );
	}

	public abstract boolean isConnected();

	public String sendCommand( String command )
		throws Exception
	{
		return sendCommand( command, 0 );
	}

	public String sendCommand( String command, int flags )
		throws Exception
	{
		logger.debug( "sendCommand: " + command );
		String result = sendCommandImpl( command, flags );
		logger.debug( "result: " + command );
		return result;
	}

	protected abstract String sendCommandImpl( String command, int flags )
		throws Exception;

	static class TelnetTerminalSession
		extends TerminalSession
	{
		private static final Logger logger = Logger.getLogger( TelnetTerminalSession.class );

		private final String host;
		private final int port;

		private TelnetSession session;

		public TelnetTerminalSession( String host, int port )
		{
			this.host = host;

			if( port > 0 )
			{
				this.port = port;
			}
			else
			{
				this.port = 23;
			}
		}

		@Override
		public void setEndSequence( String endSequence )
		{
			if( endSequence == null )
			{
				endSequence = "#";
			}

			super.setEndSequence( endSequence );

			if( session != null )
			{
				session.setEndString( endSequence );
			}
		}

		@Override
		public void connect( String username, String password )
			throws Exception
		{
			if( this.session != null )
			{
				logger.error( "Already connected!" );
				return;
			}

			TelnetSession session = new TelnetSession( host, port );

			session.setTimeout( (int)timeout );
			session.setReadWait( readWait );

			session.setEndString( ":" );

			session.connect();
			logger.info( "Connected" );

			this.session = session;

			logger.info( session.doCommand( username ) );
			logger.info( "Login entered" );

			if( endSequence == null )
			{
				endSequence = "#";
			}

			session.setEndString( endSequence );

			logger.info( session.doCommand( password ) );
			logger.info( "Password entered" );
		}

		@Override
		protected String sendCommandImpl( String command, int flags )
			throws Exception
		{
			if( flags == NOWAIT )
			{
				session.doCommandAsync( command );
				return null;
			}
			else
			{
				return session.doCommand( command );
			}
		}

		@Override
		public void close()
			throws Exception
		{
			if( session != null )
			{
				try
				{
					if( Utils.notBlankString( this.exitCommand ) )
					{
						logger.info( "sendCommandA: " + this.exitCommand );
						session.doCommandAsync( this.exitCommand );
					}
				}
				finally
				{
					session.disconnect();

					session = null;

					logger.debug( "Disconnected" );
				}
			}
		}

		@Override
		public boolean isConnected()
		{
			return session != null;
		}
	}

	static class SshTerminalSession
		extends TerminalSession
	{
		private static final Logger logger = Logger.getLogger( SshTerminalSession.class );
		
		private String privateKeyFile;

		private final String host;
		private final int port;

		private SSHSession session;

		public SshTerminalSession( String host, int port )
		{
			this.host = host;

			if( port > 0 )
			{
				this.port = port;
			}
			else
			{
				this.port = 22;
			}
        }

        @Override
        public void setConfiguration( ParameterMap config )
        {
            if( config == null )
            {
                logger.info( "setConfiguration: config is null" );
                return;
            }

            this.privateKeyFile = config.get( "privateKeyFile", null );
        }

        @Override
		public void setEndSequence( String endSequence )
		{
			if( endSequence == null )
			{
				endSequence = "#";
			}

			super.setEndSequence( endSequence );

			if( session != null )
			{
				if( endSequence.startsWith( "\\" ) && endSequence.endsWith( "\\" ) )
				{
					session.setRegexp( endSequence.substring( 1, endSequence.length() - 1 ) );
				}
				else
				{
					session.setEndString( endSequence );
				}
			}
		}

		@Override
		public void connect( String username, String password )
			throws Exception
		{
			SSHSession session = new SSHSession( host, port, username, password );
			session.setTimeout( (int)timeout );
			session.setReadWait( readWait );
			session.setPrivateKeyFile( privateKeyFile );

			if( Utils.notBlankString( endSequence ) )
			{
				session.setEndString( endSequence );
			}

			session.connect();
			logger.info( "Connected" );

			this.session = session;
		}

		@Override
		protected String sendCommandImpl( String command, int flags )
			throws Exception
		{
			if( flags == NOWAIT )
			{
				session.doCommandAsync( command );
				return null;
			}
			else
			{
				return session.doCommand( command );
			}
		}

		@Override
		public void close()
			throws Exception
		{
			if( session != null )
			{
				try
				{
					logger.info( "sendCommandA: " + this.exitCommand );
					session.doCommandAsync( this.exitCommand );
				}
				finally
				{
					session.disconnect();

					session = null;

					logger.debug( "Disconnected" );
				}
			}
		}

		@Override
		public boolean isConnected()
		{
			return session != null;
		}
	}

	static class ExecTerminalSession
		extends TerminalSession
	{
		private boolean connected = false;

		public ExecTerminalSession()
		{}

		@Override
		public void connect( String username, String password )
			throws Exception
		{
			this.connected = true;
		}

		@Override
		protected String sendCommandImpl( String command, int flags )
			throws Exception
		{
			Process process = Runtime.getRuntime().exec( command );

			if( flags == NOWAIT )
			{
				return null;
			}

			if( !process.waitFor( 1, TimeUnit.MINUTES ) )
			{
				throw new TimeoutException( "Timeout waiting when executing command." );
			}

			final ByteArrayOutputStream baos = new ByteArrayOutputStream();

			IOUtils.transfer( process.getInputStream(), baos, 256 );

			byte[] resultBytes = baos.toByteArray();

			String result = new String( resultBytes, "UTF-8" );

			return result;
		}

		@Override
		public void close()
			throws Exception
		{
			this.connected = false;
		}

		@Override
		public boolean isConnected()
		{
			return this.connected;
		}
	}

	static class TcpTerminalSession
		extends TerminalSession
	{
		private static final Logger logger = Logger.getLogger( TcpTerminalSession.class );

		private final String host;
		private final int port;

		private Socket socket;
		private PrintWriter out;
		private BufferedReader in;

		private Scanner scanner;

		private Pattern pattern;

		public TcpTerminalSession( String host, int port )
		{
			this.host = host;
			this.port = port;
		}

		@Override
		public void connect( String username, String password )
			throws Exception
		{
			socket = new Socket(  );

			socket.setSoTimeout( (int)timeout );

			socket.connect( new InetSocketAddress( host, port ), (int)timeout );

			out = new PrintWriter( socket.getOutputStream(), true );

			InputStreamReader isr = new InputStreamReader( socket.getInputStream() );
			in = new BufferedReader( isr );

			scanner = new Scanner( in );

			logger.info( "Connected" );
		}

		@Override
		public void setEndSequence( String endSequence )
		{
			super.setEndSequence( endSequence );

			pattern = Pattern.compile( Pattern.quote( endSequence ), Pattern.DOTALL );
		}

		@Override
		protected String sendCommandImpl( String command, int flags )
			throws Exception
		{
			out.println( command );
			out.flush();

			if( flags == TerminalSession.NOWAIT || Utils.isEmptyString( endSequence ) )
			{
				return "";
			}

			scanner.useDelimiter( pattern );

			return scanner.next();
		}

		@Override
		public void close()
			throws Exception
		{
			if( out != null )
			{
				out.close();
				socket.close();

				scanner.close();

				logger.info( "Disconnected" );

				out = null;
				socket = null;
			}
		}

		@Override
		public boolean isConnected()
		{
			return out != null;
		}
	}
	
	static class MikrotikTerminalSession
		extends TerminalSession
	{
		private static final Logger logger = Logger.getLogger( TelnetTerminalSession.class );

		private final String host;
		private final int port;

		private MikrotikApiSession session;

		public MikrotikTerminalSession( String host, int port )
		{
			this.host = host;

			if( port > 0 )
			{
				this.port = port;
			}
			else
			{
				this.port = 8728;
			}
		}

		@Override
		public void connect( String username, String password )
			throws Exception
		{
			if( this.session != null )
			{
				logger.error( "Already connected!" );
				return;
			}

			MikrotikApiSession session = new MikrotikApiSession( host, port, username, password, super.sourceHost, super.sourcePort );

			session.setTimeout( (int)timeout );

			session.connect();
			logger.info( "Connected" );

			this.session = session;
		}

		@Override
		protected String sendCommandImpl( String command, int flags )
			throws Exception
		{
			return session.doCommand( command.replace( "\\n", "\n" ) );
		}

		@Override
		public void close()
			throws Exception
		{
			if( session != null )
			{
				try
				{
					if( Utils.notBlankString( this.exitCommand ) )
					{
						logger.info( "sendCommand: " + this.exitCommand );
						session.doCommand( this.exitCommand );
					}
				}
				finally
				{
					session.disconnect();

					session = null;

					logger.debug( "Disconnected" );
				}
			}
		}

		@Override
		public boolean isConnected()
		{
			return session != null;
		}
	}

	public static TerminalSession newSshSession( String host, int port )
	{
		return new SshTerminalSession( host, port );
	}

	public static TerminalSession newTelnetSession( String host, int port )
	{
		return new TelnetTerminalSession( host, port );
	}

	public static TerminalSession newTcpSession( String host, int port )
	{
		return new TcpTerminalSession( host, port );
	}

	public static TerminalSession newTerminalSession( String type, String host, int port )
	{
		return newTerminalSession( type, host, port, null, null );
	}

	public static TerminalSession newTerminalSession( String type, String host, int port, String username, String password )
	{
		TerminalSession result;

		switch( type )
		{
			case "ssh":
				result = new SshTerminalSession( host, port );
				break;

			case "shell":
				result = new ExecTerminalSession();
				break;

			case "tcp":
				result = new TcpTerminalSession( host, port );
				break;
				
			case "mikrotik":
				result = new MikrotikTerminalSession( host, port );
				break;

			case "telnet":
			default:
				result = new TelnetTerminalSession( host, port );
				break;
		}

		result.setUsername( username );
		result.setPassword( password );

		return result;
	}

	public static TerminalSession newTerminalSession( URI uri )
	{
		return newTerminalSession( uri.getScheme(), uri.getHost(), uri.getPort() );
	}

	public static TerminalSession newTerminalSession( Device<?, ?> device, ParameterMap config, String prefix )
	{
		String host = null;
		int port = 0;

		if( device != null )
		{
			List<InetSocketAddress> hosts = device.getHosts();
			if( hosts != null && hosts.size() > 0 )
			{
				InetSocketAddress socketAddress = hosts.get( 0 );
				host = socketAddress.getAddress().getHostAddress();
				port = socketAddress.getPort();
			}
		}

		host = config.get( prefix + "host", host );
		port = config.getInt( prefix + "port", port );

		String type = config.get( prefix + "protocol", "telnet" );

		if( port <= 0 )
		{
			if( "telnet".equals( type ) )
			{
				port = 23;
			}
			else if( "ssh".equals( type ) )
			{
				port = 22;
			}
		}

		String username;
		String password;

		if( device != null )
		{
			username = device.getUsername();
			password = device.getPassword();
		}

		username = config.get( prefix + "username", device.getUsername() );
		password = config.get( prefix + "password", device.getPassword() );

		TerminalSession result = TerminalSession.newTerminalSession( type, host, port, username, password );
		
		result.setSourceHost( config.get( prefix + "sourceHost", null ) );
		result.setSourcePort( config.getInt( prefix + "sourcePort", 0 ) );

		String endSequence = config.get( prefix + "endSequence", null );

		result.setEndSequence( endSequence );

		return result;
	}

	/**
	 * Дополнительная конфигурация.
	 * @param config
	 */
    public void setConfiguration( ParameterMap config )
    {
    }
}
